@@ -3,6 +3,7 @@ import os
 import re
 import time
 import traceback
+from datetime import datetime, timedelta
 from pathlib import Path
 from typing import Optional
 
@@ -19,16 +20,24 @@ from webdriver_manager.firefox import GeckoDriverManager
 
 logger = logging.getLogger("BibQuery")
 
+# Disable annoying webdriver manager outputs
+os.environ["WDM_PROGRESS_BAR"] = "0"
+
 
 class BibQueryException(Exception):
     pass
 
 
+class CaptchaEncounteredException(Exception):
+    pass
+
+
 class BibQuery:
     def __init__(self):
         self.__browser: Optional[WebDriver] = None
         self.__res_path = Path(__file__).parent / "res"
         self.__cache_path = Path("~").expanduser() / ".cache" / "bibquery"
+        self.__cookie_path = self.__cache_path / "google_cookies.json"
         with (self.__res_path / "urlSpecificAdjusterList.json").open() as f:
             self.__url_specific_adjusters = json.load(f)
 
@@ -43,9 +52,28 @@ class BibQuery:
         options = Options()
         options.headless = True
         self.__cache_path.mkdir(exist_ok=True, parents=True)
-        self.__browser = webdriver.Firefox(executable_path=GeckoDriverManager(
-            path=str(self.__cache_path)).install(), options=options, log_path=os.devnull)
+
+        # Workaround for the GitHub rate limit issue (https://github.com/SergeyPirogov/webdriver_manager/issues/442)
+        with (self.__cache_path / ".wdm" / "drivers.json").open() as f:
+            drivers_dict = json.load(f)
+        max_cache_age = timedelta(days=1)
+        driver_filename = None
+        for name, properties in drivers_dict.items():
+            timestamp = datetime.strptime(properties["timestamp"], "%d/%m/%Y")
+            if timestamp - datetime.today() < max_cache_age:
+                driver_filename = properties["binary_path"]
+        if driver_filename is None:
+            driver_filename = GeckoDriverManager(path=str(self.__cache_path)).install()
+
+        self.__browser = webdriver.Firefox(executable_path=driver_filename, options=options, log_path=os.devnull)
         self.__browser.install_addon(self.__res_path / "bibitnow_patched.xpi", temporary=True)
+        if self.__cookie_path.exists():
+            with self.__cookie_path.open() as f:
+                cookies = json.load(f)
+            self.__browser.get("https://scholar.google.com/")
+            self.__browser.delete_all_cookies()
+            for cookie in cookies:
+                self.__browser.add_cookie(cookie)
 
     def close(self):
         self.__browser.close()
@@ -88,6 +116,8 @@ class BibQuery:
         domain = url_parsed.netloc.lower()
         *rest, top = domain.split(".")
         rest_joined = ".".join(rest)
+        if rest_joined.startswith("www."):
+            rest_joined = rest_joined[4:]
         path = url_parsed.path[1:]
         for adjuster in self.__url_specific_adjusters:
             if re.match(adjuster["scheme"], rest_joined):
@@ -159,13 +189,50 @@ class BibQuery:
         if self.__browser is None:
             raise ValueError("BibQuery has not been initialized or was already closed.")
 
-        self.__browser.get("https://scholar.google.com/?")
-        self.__browser.find_element(By.NAME, "q").send_keys(url)
-        self.__browser.find_element(By.NAME, "btnG").click()
+        try:
+            return self.__query_google_scholar(url, self.__browser)
+        except CaptchaEncounteredException:
+            logger.info("Encountered Captcha. Prompting user to solve it.")
+            browser = webdriver.Firefox(executable_path=GeckoDriverManager(
+                path=str(self.__cache_path), cache_valid_range=14).install(), log_path=os.devnull)
+            try:
+                return_value = self.__query_google_scholar(url, browser, cancel_on_captcha=False)
 
-        results_element = self.__wait_and_get(self.__browser, By.XPATH, "//div[@id='gs_res_ccl_mid']")
-        results_element.find_element(By.XPATH, "//a[@aria-controls='gs_cit']").click()
-        link = self.__wait_and_get(self.__browser, By.XPATH, "//a[contains(text(), 'BibTeX')]").get_attribute("href")
+                # This seems to be the only way of obtaining and setting cookies for www.google.com
+                browser.get("https://scholar.google.com/")
+                cookies = browser.get_cookies()
+                with self.__cookie_path.open("w") as f:
+                    json.dump(cookies, f)
+                self.__browser.get("https://scholar.google.com/")
+                self.__browser.delete_all_cookies()
+                for cookie in cookies:
+                    self.__browser.add_cookie(cookie)
+            finally:
+                browser.close()
+            return return_value
 
-        self.__browser.get(link)
-        return self.__browser.find_element(By.XPATH, "/html/body/pre").text
+    def __query_google_scholar(self, url: str, browser: WebDriver, cancel_on_captcha: bool = True) -> str:
+        browser.get("https://scholar.google.com/")
+        browser.find_element(By.NAME, "q").send_keys(url)
+        browser.find_element(By.NAME, "btnG").click()
+
+        start_time = time.time()
+        citation_element = None
+        while (time.time() - start_time < 60.0 or not cancel_on_captcha) and citation_element is None:
+            citation_elements = browser.find_elements(By.XPATH, "//a[@aria-controls='gs_cit']")
+            if len(citation_elements) > 0:
+                citation_element = citation_elements[0]
+            else:
+                recaptcha_element_present = \
+                    len(browser.find_elements(By.XPATH, "//iframe[@title='reCAPTCHA']")) > 0
+                if recaptcha_element_present and cancel_on_captcha:
+                    raise CaptchaEncounteredException("Encountered Captcha when querying Google Scholar.")
+                time.sleep(0.1)
+        if citation_element is None:
+            raise TimeoutError("Timed out waiting for the citation link to appear on Google Scholar.")
+        citation_element.click()
+
+        link = self.__wait_and_get(browser, By.XPATH, "//a[contains(text(), 'BibTeX')]").get_attribute("href")
+
+        browser.get(link)
+        return browser.find_element(By.XPATH, "/html/body/pre").text
